import os
from unittest.mock import call, patch

from django.db import connections
from django.test import TestCase

from test_queries import NumQueriesMixin

current_directory = os.getcwd()


class TestNumQueriesMixin(NumQueriesMixin, TestCase):
    def test_basic_functionality(self):
        with self.assertNumQueries(1):
            connections["default"].cursor().execute("SELECT 1")

    def test_file_creation(self):
        # This test checks if the SQL log file is created
        # You might need to adjust the expected filename based on your implementation
        expected_filename = "path_to_expected_file.sqllog"  # Adjust this
        with self.assertNumQueries(1):
            connections["default"].cursor().execute("SELECT 1")
        self.assertTrue(os.path.exists(expected_filename))

    def test_file_comparison(self):
        # This test checks if the SQL log files are compared correctly
        # You might need to adjust the expected filename and contents based on your implementation
        expected_filename = "path_to_expected_file.sqllog"  # Adjust this
        with open(expected_filename, "w") as f:
            f.write("SELECT 1")
        with self.assertNumQueries(1):
            connections["default"].cursor().execute("SELECT 1")

    def test_file_comparison_function(self):
        expected_filename = "path_to_expected_file.sqllog"  # Adjust this
        with open(expected_filename, "w") as f:
            f.write("SELECT 1")

        def call_sql():
            connections["default"].cursor().execute("SELECT 1")

        self.assertNumQueries(1, call_sql)

    def test_environment_variables(self):
        os.environ["TEST_QUERIES_DISABLE"] = "1"
        with self.assertNumQueries(1):
            connections["default"].cursor().execute("SELECT 1")
        # Test content of the SQL log files
        with open(
            "tests/sqllog/test_num_queries.TestNumQueriesMixin.test_environment_variables.0.sqllog.new",
            "r",
        ) as f:
            self.assertEqual(f.read(), "SELECT 1\n")

    def test_file_comparison_with_existing_lines(self):
        # This test checks if the SQL log files are compared correctly when the file already has lines
        # You might need to adjust the expected filename and contents based on your implementation
        expected_filename = (
            "tests/sqllog/test_num_queries.TestNumQueriesMixin.test_file_comparison_with_existing_lines.0.sqllog"
        )
        with open(expected_filename, "w") as f:
            f.write("SELECT 1\nSELECT 2")  # Writing multiple lines to the file
        with self.assertNumQueries(2):
            connections["default"].cursor().execute("SELECT 1")
            connections["default"].cursor().execute("SELECT 2")
        # Test content of the SQL log files
        with open(
            "tests/sqllog/test_num_queries.TestNumQueriesMixin.test_file_comparison_with_existing_lines.0.sqllog.new",
            "r",
        ) as f:
            self.assertEqual(f.read(), "SELECT 1\nSELECT 2\n")

    def test_file_comparison_with_existing_lines_not_equal(self):
        expected_filename = (
            "tests/sqllog/test_num_queries.TestNumQueriesMixin.test_file_comparison_with_existing_lines.0.sqllog"
        )
        with open(expected_filename, "w") as f:
            f.write("SELECT 1")  # Writing multiple lines to the file
        with patch("builtins.print") as mock_print:
            with self.assertRaisesRegex(AssertionError, "Captured queries were:\n1. SELECT 1\n2. SELECT 2"):
                with self.assertNumQueries(1):
                    connections["default"].cursor().execute("SELECT 1")
                    connections["default"].cursor().execute("SELECT 2")
            self.assertEquals(mock_print.call_args_list[0], call("New query was recorded:"))
            self.assertIn("sql: SELECT 1", mock_print.mock_calls[1].args[0]),
            self.assertEquals(mock_print.call_args_list[2], call("See difference:"))
            self.assertEquals(
                mock_print.call_args_list[3],
                call(
                    f"  diff {current_directory}/tests/sqllog/test_num_queries."
                    "TestNumQueriesMixin.test_file_comparison_with_existing_lines_not_equal.0.sqllog "
                    f"{current_directory}/tests/sqllog/test_num_queries."
                    "TestNumQueriesMixin.test_file_comparison_with_existing_lines_not_equal.0.sqllog.new"
                ),
            )
            self.assertEquals(len(mock_print.call_args_list), 4)

    def test_file_comparison_with_existing_lines_delete(self):
        expected_filename = (
            "tests/sqllog/test_num_queries."
            "TestNumQueriesMixin.test_file_comparison_with_existing_lines_delete.0.sqllog"
        )
        with open(expected_filename, "w") as f:
            f.write("SELECT 1\nSELECT 3\nSELECT 3")  # Writing multiple lines to the file
        with patch("builtins.print") as mock_print:
            with self.assertRaisesRegex(AssertionError, "Captured queries were:\n1. SELECT 1\n2. SELECT 2"):
                with self.assertNumQueries(1):
                    connections["default"].cursor().execute("SELECT 1")
                    connections["default"].cursor().execute("SELECT 2")
                    connections["default"].cursor().execute("SELECT 3")
            self.assertEquals(mock_print.call_args_list[0], call("New query was recorded:"))
            self.assertIn("sql: SELECT 2", mock_print.mock_calls[1].args[0]),
            self.assertEquals(mock_print.call_args_list[2], call("See difference:"))
            self.assertEquals(
                mock_print.call_args_list[3],
                call(
                    f"  diff {current_directory}/tests/sqllog/test_num_queries."
                    "TestNumQueriesMixin.test_file_comparison_with_existing_lines_delete.0.sqllog "
                    f"{current_directory}/tests/sqllog/test_num_queries."
                    "TestNumQueriesMixin.test_file_comparison_with_existing_lines_delete.0.sqllog.new"
                ),
            )
            self.assertEquals(
                mock_print.call_args_list[4],
                call("delete    a[2:3] --> b[3:3] ['SELECT 3...'] --> []"),
            )
            self.assertEquals(len(mock_print.call_args_list), 5)

    def test_file_comparison_with_existing_lines_replace(self):
        expected_filename = (
            "tests/sqllog/test_num_queries."
            "TestNumQueriesMixin.test_file_comparison_with_existing_lines_replace.0.sqllog"
        )
        with open(expected_filename, "w") as f:
            f.write("SELECT 1\nSELECT 2\nSELECT 3\nSELECT 4\nSELECT 5")  # Writing multiple lines to the file
        with patch("builtins.print") as mock_print:
            with self.assertRaisesRegex(AssertionError, "Captured queries were:\n1. SELECT 1\n2. SELECT 2"):
                with self.assertNumQueries(1):
                    connections["default"].cursor().execute("SELECT 1")
                    connections["default"].cursor().execute("SELECT 2")
                    connections["default"].cursor().execute("SELECT 0")
                    connections["default"].cursor().execute("SELECT 4")
                    connections["default"].cursor().execute("SELECT 5")
            self.assertEquals(mock_print.call_args_list[0], call("Query was replaced:"))
            self.assertIn("sql: SELECT 0", mock_print.mock_calls[1].args[0]),
            self.assertEquals(mock_print.call_args_list[2], call("See difference:"))
            self.assertEquals(
                mock_print.call_args_list[3],
                call(
                    f"  diff {current_directory}/tests/sqllog/test_num_queries."
                    "TestNumQueriesMixin.test_file_comparison_with_existing_lines_replace.0.sqllog "
                    f"{current_directory}/tests/sqllog/test_num_queries."
                    "TestNumQueriesMixin.test_file_comparison_with_existing_lines_replace.0.sqllog.new"
                ),
            )
            self.assertEquals(len(mock_print.call_args_list), 4)



import os
import webbrowser

from invoke import task


def open_browser(path):
    try:
        from urllib import pathname2url
    except ImportError:
        from urllib.request import pathname2url
    webbrowser.open("file://" + pathname2url(os.path.abspath(path)))


@task
def clean_build(c):
    """
    Remove build artifacts
    """
    c.run("rm -fr build/")
    c.run("rm -fr dist/")
    c.run("rm -fr *.egg-info")


@task
def clean_pyc(c):
    """
    Remove python file artifacts
    """
    c.run("find . -name '*.pyc' -exec rm -f {} +")
    c.run("find . -name '*.pyo' -exec rm -f {} +")
    c.run("find . -name '*~' -exec rm -f {} +")


@task
def coverage(c):
    """
    check code coverage quickly with the default Python
    """
    c.run("coverage run --source dj-test-queries runtests.py tests")
    c.run("coverage report -m")
    c.run("coverage html")
    c.run("open htmlcov/index.html")


@task
def docs(c):
    """
    Build the documentation and open it in the browser
    """
    c.run("rm -f docs/dj-test-queries.rst")
    c.run("rm -f docs/modules.rst")
    c.run("sphinx-apidoc -o docs/ test_queries")

    c.run("sphinx-build -E -b html docs docs/_build")
    open_browser(path="docs/_build/html/index.html")


@task
def test_all(c):
    """
    Run tests on every python version with tox
    """
    c.run("tox")


@task
def clean(c):
    """
    Remove python file and build artifacts
    """
    clean_build(c)
    clean_pyc(c)


@task
def unittest(c):
    """
    Run unittests
    """
    c.run("python manage.py test")


@task
def lint(c):
    """
    Check style with flake8
    """
    c.run("flake8 dj-test-queries tests")


@task(help={"bumpsize": 'Bump either for a "feature" or "breaking" change'})
def release(c, bumpsize=""):
    """
    Package and upload a release
    """
    clean(c)
    if bumpsize:
        bumpsize = "--" + bumpsize

    c.run("bumpversion {bump} --no-input".format(bump=bumpsize))

    import dj_test_queries

    c.run("python setup.py sdist bdist_wheel")
    c.run("twine upload dist/*")

    c.run('git tag -a {version} -m "New version: {version}"'.format(version=dj_test_queries.__version__))
    c.run("git push --tags")
    c.run("git push origin master")



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from django.contrib import admin
from django.urls import path

urlpatterns = [
    path("admin/", admin.site.urls),
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    """Retrieves the version from test_queries/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


version = get_version("test_queries", "__init__.py")


if sys.argv[-1] == "publish":
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    sys.exit()

if sys.argv[-1] == "tag":
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
readme = open("README.rst").read()
history = open("HISTORY.rst").read().replace(".. :changelog:", "")
requirements = open(os.path.join(BASE_DIR, "requirements.txt")).readlines()

setup(
    name="dj-test-queries",
    version=version,
    description=(
        "Extension of assertNumQueries that can record queries "
        "that were executed and show the differences in future runs."
    ),
    long_description=readme + "\n\n" + history,
    author="Petr Dlouhý",
    author_email="petr.dlouhy@email.cz",
    url="https://github.com/PetrDlouhy/dj-test-queries",
    packages=[
        "test_queries",
    ],
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords="dj-test-queries",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)



import inspect
import os
import re
from difflib import SequenceMatcher
from typing import Any, Dict

from debug_toolbar.panels.sql.tracking import wrap_cursor
from django.db import DEFAULT_DB_ALIAS, connections
from django.test import TransactionTestCase
from django.test.testcases import _AssertNumQueriesContext


class Logger(object):
    def __init__(self, context):
        self.queries = []
        self.context = context

    def record(self, sql, raw_sql, stacktrace, *args, **kwargs):
        self.context["records"].append(
            {
                "sql": sql,
                "raw_sql": re.sub(r'SAVEPOINT ".*"', 'SAVEPOINT "%1"', raw_sql),
                "stacktrace": stacktrace,
                "args": args,
                **kwargs,
            }
        )

    def current_transaction_id(self, alias):
        pass

    def new_transaction_id(self, alias):
        pass


def get_error_string(error_dict):
    string = ""
    string += error_dict["raw_sql"]
    string += "\n"
    string += "Stacktrace:\n"
    string += "\n".join(f'  File: "{s[0]}", Line: {s[1]}, in {s[2]}\n    {s[3]}' for s in error_dict["stacktrace"])
    string += "\n"
    string += "\n".join([f"{rk}: {rv}" for rk, rv in error_dict.items() if rk in ["sql", "params"]])
    string += "\n"
    string += "-----------------------------------------"
    return string


class _AssertQueriesContext(_AssertNumQueriesContext):
    def __init__(self, test_case, num, connection, context_dict):
        self.context_dict = context_dict
        super().__init__(test_case, num, connection)

    def __exit__(self, exc_type, exc_value, traceback):
        filename = self.context_dict["filename"]
        try:
            with open(filename, "r") as f:
                lines = f.readlines()
        except FileNotFoundError:
            lines = None
        raw_queries = [s["raw_sql"] for s in self.context_dict["records"]]
        if lines:
            lines = [line.strip() for line in lines]
            s = SequenceMatcher(None, lines, raw_queries)
            for tag, i1, i2, j1, j2 in s.get_opcodes():
                if tag not in ["insert", "replace", "equal"]:
                    print(
                        "{:7}   a[{}:{}] --> b[{}:{}] {!r:>8} --> {!r}".format(
                            tag,
                            i1,
                            i2,
                            j1,
                            j2,
                            [s[:50] + "..." for s in lines[i1:i2]],
                            [s[:50] + "..." for s in raw_queries[j1:j2]],
                        )
                    )
                if tag in ["insert", "replace"]:
                    for j in range(j1, j2):
                        if tag == "insert":
                            print("New query was recorded:")
                        elif tag == "replace":
                            print("Query was replaced:")
                        print(get_error_string(self.context_dict["records"][j]))
                        print("See difference:")
                        print(f"  diff {filename} {filename}.new")

        if not os.environ.get("TEST_QUERIES_REWRITE_SQLLOGS"):
            filename += ".new"
        os.makedirs(filename.rsplit("/", 1)[0], exist_ok=True)
        with open(filename, "w") as f:
            f.write("\n".join(raw_queries) + "\n")

        return super().__exit__(exc_type, exc_value, traceback)


class NumQueriesMixin(TransactionTestCase):
    executed_times = 0
    context: Dict[str, Any] = {}

    def assertNumQueries(self, num, func=None, *args, using=DEFAULT_DB_ALIAS, **kwargs):  # noqa: N802
        conn = connections[using]
        path, file_prefix = inspect.getmodule(self).__file__[:-3].rsplit("/", 1)
        filename = (
            f"{path}/sqllog/{file_prefix}."
            f"{self.__class__.__name__}."
            f"{self._testMethodName}.{self.executed_times}.sqllog"
        )
        self.executed_times += 1
        self.context["filename"] = filename
        self.context["records"] = []
        logger = Logger(context=self.context)
        conn._djdt_logger = logger

        try:  # DDT >= 4.2.0
            wrap_cursor(conn)
        except TypeError:
            wrap_cursor(conn, logger)
        context = _AssertQueriesContext(self, num, conn, self.context)

        if func is None:
            return context

        with context:
            func(*args, **kwargs)
